VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "clsTrueRnd"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit
Option Compare Text

' Used in waiting
Private Declare Sub Sleep Lib "kernel32" (ByVal dwMilliseconds As Long)

' The different limits
Private Type InputLimit
    Min As Long
    Max As Long
End Type

' Property containing a service
Private Type Service
    Name As String
    Description As String
    RequestType As String
    RequestURL As String
    RequestAsync As Boolean
    DataString As String
    ReadBegin As String
    ReadEnd As String
    ReadDelimiter As String
    ReadReplace As String
    ReadMethod As Long
    ReadTrim As Boolean
    AmountLimit As InputLimit
    NumberLimit As InputLimit
End Type

' The lenght of the downloaded buffer (amount)
Public BufferLenght As Long

' How much accuracy which is needed
Public BufferAccuracy As Long

' How much of the buffer that has to be used until it gets refreshed
Public BufferReload As Double

' The current selected service
Public CurrentService As Long

' The class used to contact and download from the HTTP server
Private RequestXML As New XMLHTTP

' The array containing all the services
Private aServices() As Service
Private lngCount As Long

' The array containing all the random numbers (entropy pool)
Private aNumbers As Variant
Private NumberPos As Long
Private BufferUnderway As Boolean

' Will append the content of a XML-file
Public Sub LoadServices(sXMLPath As String)

    Dim Document As New DOMDocument, Service As IXMLDOMElement, Request As IXMLDOMElement
    Dim Limits As IXMLDOMElement, Reader As IXMLDOMElement
    
    ' Open the given XML-document
    Document.Load sXMLPath
    
    ' Go through all the services
    For Each Service In FindChild(Document, "collection").childNodes
    
        ' Retrieve different subelements
        Set Request = FindChild(Service, "request")
        Set Limits = FindChild(Service, "limits")
        Set Reader = FindChild(Service, "reader")
    
        ' Increase the amount of services
        lngCount = lngCount + 1
        
        ' Reallocate array
        ReDim Preserve aServices(1 To lngCount)

        ' Fill the service post with information
        With aServices(lngCount)
            .Name = Service.getAttribute("name")
            .RequestType = Request.getAttribute("type")
            .RequestAsync = (Request.getAttribute("async") = "true")
            .RequestURL = Request.Text
            .Description = FindChild(Service, "description").Text
            .DataString = FindChild(Service, "data").Text
            .AmountLimit.Max = FindChild(Limits, "amount").getAttribute("max")
            .AmountLimit.Min = FindChild(Limits, "amount").getAttribute("min")
            .NumberLimit.Max = FindChild(Limits, "number").getAttribute("max")
            .NumberLimit.Min = FindChild(Limits, "number").getAttribute("min")
            .ReadBegin = FindChild(Reader, "begin").Text
            .ReadEnd = FindChild(Reader, "end").Text
            .ReadReplace = FindChild(Reader, "replace").Text
            .ReadDelimiter = FindChild(Reader, "delimiter").Text
            .ReadMethod = Val(FindChild(Reader, "method").Text)
            .ReadTrim = (FindChild(Reader, "trim").Text = "true")
        End With
    
    Next

End Sub

Private Function FindChild(oParent As Object, sName As String) As IXMLDOMElement

    ' Simply return the first element found
    Set FindChild = oParent.getElementsByTagName(sName)(0)

End Function

Public Sub SelectService(ByVal Index As Long)

    ' If this is a valid index, ...
    If IsService(Index) Then
        ' ... select this service
        CurrentService = Index
    End If

End Sub

Public Property Get ServiceCount() As Long

    ' Return the amount of services registered
    ServiceCount = lngCount

End Property

Public Property Get ServiceName(ByVal Index As Long) As String
    
    ' Return the name of the service
    If IsService(Index) Then
        ServiceName = aServices(Index).Name
    End If

End Property

Public Property Get ServiceDescription(ByVal Index As Long) As String
    
    ' Return the description of the service
    If IsService(Index) Then
        ServiceDescription = aServices(Index).Description
    End If

End Property

Public Property Get ServiceMax(ByVal Index As Long) As Long
    
    ' Return the maximum amount of numbers to download for this service
    If IsService(Index) Then
        ServiceMax = aServices(Index).AmountLimit.Max
    End If

End Property

Public Function IsService(ByVal Index As Long) As Boolean

    ' Return whether or not this is a valid index
    IsService = CBool(Index > 0 And Index <= lngCount)

End Function

' Will only return a number within the given limit
Private Function KeepWithin(ByVal Value As Long, Limit As InputLimit) As Long

    ' Make sure the value is within the limit
    If Value < Limit.Min Then
        KeepWithin = Limit.Min
    ElseIf Value > Limit.Max Then
        KeepWithin = Limit.Max
    Else
        KeepWithin = Value
    End If

End Function

' Used to reinitialize and download fresh random numbers from the current service
Public Sub UpdateBuffer(Optional overrideAsync As Variant)
Attribute UpdateBuffer.VB_UserMemId = 0

    Dim bAsync As Boolean, sData As String

    ' See if a transmission has been underway
    If BufferUnderway Then
    
        ' See if the transmission is complete
        If RequestXML.readyState = 4 Then
        
            ' Process the result
            If aServices(CurrentService).ReadMethod <> 2 Then
                ProcessResult RequestXML.responseText
            Else
                aNumbers = RequestXML.responseBody
                NumberPos = LBound(aNumbers)
            End If
            
            ' Clear this buffer
            Set RequestXML = New XMLHTTP
            
            ' The buffer has been sent
            BufferUnderway = False
            
        End If
    
        ' Don't continue
        Exit Sub
    
    End If

    ' Use the current selected service
    With aServices(CurrentService)
    
        ' Reset the event handler
        RequestXML.onreadystatechange = Nothing
        
        ' Make sure the buffer lenght is valid
        BufferLenght = KeepWithin(BufferLenght, .AmountLimit)
        
        ' See if we are to override a property or not
        If IsMissing(overrideAsync) Then
            bAsync = .RequestAsync
        Else
            bAsync = CBool(overrideAsync)
        End If
        
        ' Process the data-string
        sData = ProcessString(.DataString, "amount", BufferLenght, _
                 "lowerlimit", .NumberLimit.Min, "upperlimit", .NumberLimit.Max)
        
        ' Request the HTTP-site for random numbers
        Select Case .RequestType
        Case "get" ' Use the HTTP get command
            RequestXML.open .RequestType, .RequestURL & "?" & sData, bAsync
            RequestXML.send ""
            
        Case "post" ' Use the HTTP post command
            RequestXML.open .RequestType, .RequestURL, bAsync
            RequestXML.send sData
        End Select
    
        ' If we're not requesting the page asynchronously, we can parse the result right away
        If bAsync = False Then
        
            ' Just process the result
            If .ReadMethod <> 2 Then
                ProcessResult RequestXML.responseText
            Else
                aNumbers = RequestXML.responseBody
                NumberPos = LBound(aNumbers)
            End If
            
            ' Clear this buffer
            Set RequestXML = New XMLHTTP
            
        Else
            
            ' Use the current module
            RequestXML.onreadystatechange = Me
        
            ' Let us wait for a response
            BufferUnderway = True
        
        End If
    
    End With
    
End Sub

' Returns a random number within 0 and 1
Public Function RandomNumber() As Double

    ' Make sure there's enough bytes to execute the request
    Do Until (NumberPos + BufferAccuracy - 1 < BufferLenght)
        DoEvents
        Sleep 10
    Loop
 
    ' Calculate a random number (use several values to increase accuracy)
    With aServices(CurrentService)
        RandomNumber = (SumNumbers(aNumbers, NumberPos, BufferAccuracy) - (.NumberLimit.Min * _
         BufferAccuracy)) / (.NumberLimit.Max - .NumberLimit.Min) / BufferAccuracy
    End With
    
    ' Increase the read-position
    NumberPos = NumberPos + BufferAccuracy
    
    ' If we're dangerously reaching the end, automatically get new numbers
    If BufferUsed > BufferReload Then
        UpdateBuffer
    End If

End Function

Public Function PeekArray() As Variant

    ' Return the entire buffer
    PeekArray = aNumbers

End Function

' Summerize values in an array
Private Function SumNumbers(aArray, ByVal StartIndex As Long, ByVal Lenght As Long) As Double

    Dim Tell As Long
    
    ' Go through the specified indexes, ...
    For Tell = StartIndex To StartIndex + Lenght - 1
    
        ' ... adding them all
        SumNumbers = SumNumbers + aArray(Tell)
        
    Next

End Function

Private Function ProcessString(ByVal sText As String, ParamArray aParam() As Variant) As String

    Dim Tell As Long
    
    ' Go through all the given paramenters
    For Tell = LBound(aParam) To UBound(aParam) Step 2
    
        ' Replace the in-line variable
        sText = Replace(sText, "%" & aParam(Tell) & "%", aParam(Tell + 1))
    
    Next

    ' Return the processed string
    ProcessString = sText

End Function

' Processes a returned HTML-document with the random numbers
Private Sub ProcessResult(sHTML As String)
    
    'On Error Resume Next
    Dim posBegin As Long, posEnd As Long, sBuffer As String, bData() As Byte

    ' Use the current selected service
    With aServices(CurrentService)
    
        ' Set the default positions
        posBegin = 1
        posEnd = Len(sHTML)
    
        ' Firstly, find the positions of the start position and the end position respectly
        Select Case .ReadMethod
        Case 0 ' Use normal string search
            posBegin = InStr(1, sHTML, .ReadBegin) + Len(.ReadBegin)
            posEnd = InStr(posBegin, sHTML, .ReadEnd)
            
        Case 1 ' Use absolute positions
            If .ReadBegin > 0 Then posBegin = .ReadBegin
            If .ReadEnd > 0 Then posEnd = .ReadEnd
        End Select

        ' Retrieve the data section
        sBuffer = Mid(sHTML, posBegin, posEnd - posBegin)
        
        ' Trim it if said so
        If .ReadTrim Then
            sBuffer = Trim$(sBuffer)
        End If
        
        ' Fix the string with the replace-string
        FixString sBuffer, .ReadReplace

        ' Then, split the numbers and save it in the buffer
        aNumbers = Split(sBuffer, ConvertFromHex(.ReadDelimiter))

        ' Reset the buffer position
        NumberPos = LBound(aNumbers)
    
    End With

End Sub

' Uses a/several patterns (%00%00=%0A;%01%01=) to fix a string
Private Function FixString(refText As String, sPattern As String)

    Dim aPatterns As Variant, aCommand As Variant, Tell As Long
    
    ' Firstly, extract all the patterns
    aPatterns = Split(sPattern, ";")
    
    ' Then go through them all
    For Tell = LBound(aPatterns) To UBound(aPatterns)
    
        ' Parse the command
        aCommand = Split(aPatterns(Tell), "=", 2)
        
        ' Replace and fix the string
        refText = Replace(refText, ConvertFromHex(CStr(aCommand(0))), ConvertFromHex(CStr(aCommand(1))))
    
    Next

End Function

' Converts a string of hexadecimals to a real character string
Private Function ConvertFromHex(sHex As String)

    Dim Tell As Long, sBuffer As String
    
    ' Firstly, allocate the buffer (1/3 of the lenght of the hex)
    sBuffer = Space(Len(sHex) / 3)
    
    ' Read three and three characters
    For Tell = 1 To Len(sHex) Step 3
    
        ' Set the current character
        Mid(sBuffer, ((Tell - 1) / 3) + 1, 1) = Chr(Val("&H" & Mid(sHex, Tell + 1, 2)))
    
    Next

    ' Return the buffer
    ConvertFromHex = sBuffer

End Function

Public Property Get BufferUsed() As Double

    ' Return a number on how much of the buffer that has been used
    BufferUsed = NumberPos / BufferLenght

End Property

Private Sub Class_Initialize()

    ' Set the default values
    CurrentService = 1
    BufferAccuracy = 2
    BufferReload = 0.9

End Sub
